from .alphaquery import AlphaQuery
from .barchart import BarChart
from .exceptions import HttpException
from .finviz import FinViz
from .stockanalysis import StockAnalysis
from .stocktwits import StockTwits
from .tipranks import TipRanks

# from .yahoo import Yahoo
# from .etfdb import EtfDb



# import copy
# import re
#
# from .base import BaseAPI, HtmlTableParser, strip_html_tags
# from .exceptions import DataException, SymbolNotFound
#
#
# class EtfDb(BaseAPI):
#     """
#     EtfDb.com
#     """
#
#     def get_basic_info(self):
#         """
#         Returns basic info about the given ETF. The info is:
#
#         - Issuer
#         - Brand
#         - Structure
#         - Expense Ratio
#         - Inception
#         - Index Tracked
#         - Category
#         - Leveraged
#         - Asset Class
#         - Asset Class Size
#         - Region (General)
#         - Region (Specific)
#         - Description
#         - Assets Under Management
#
#         :return: ETF basic info as a dict.
#         :rtype: dict
#         """
#
#         # Get HTML.
#         try:
#             html = self._get(
#                 f"https://etfdb.com/etf/{self.symbol.upper()}/#etf-ticker-profile"
#             )
#         except Exception as e:
#             raise SymbolNotFound from e
#
#         finds = re.findall(
#             r"class=\'ticker-assets\'.*?>(.*?)</div>[^<]*</div>", html.text, re.DOTALL
#         )
#
#         # Check if the HTML contains only two occurrences.
#         if 0 == len(finds):
#             raise SymbolNotFound
#         if 2 < len(finds):
#             raise DataException(
#                 "More that two occurrences found in HTML - don't know what to do now"
#             )
#
#         # Process 1st list.
#         list_items = re.findall(r"<span.*?>(.*?)</span>", finds[0], re.DOTALL)
#         list_items = [strip_html_tags(i) for i in list_items]
#         data = dict(zip(list_items[::2], list_items[1::2]))
#
#         # Process 2nd list.
#         list_items = re.findall(r"<span.*?>(.*?)</span>", finds[1], re.DOTALL)
#         list_items = [strip_html_tags(i) for i in list_items]
#         data |= dict(zip(list_items[::2], list_items[1::2]))
#
#         for key in copy.deepcopy(data).keys():
#             if key not in [
#                 "Issuer",
#                 "Brand",
#                 "Structure",
#                 "Expense Ratio",
#                 "Inception",
#                 "Index Tracked",
#                 "Category",
#                 "Leveraged",
#                 "Asset Class",
#                 "Asset Class Size",
#                 "Region (General)",
#                 "Region (Specific)",
#             ]:
#                 del data[key]
#
#         # Fetch description.
#         finds = re.findall(
#             r"id='analyst-report'>.*?<p><p><p>(.+?)</p>", html.text, re.DOTALL
#         )
#
#         # Check if the HTML contains only ine occurrences.
#         if 0 == len(finds):
#             raise SymbolNotFound
#         if 1 < len(finds):
#             raise DataException(
#                 "More that one occurrences found in HTML - don't know what to do now"
#             )
#
#         data["Description"] = strip_html_tags(finds[0])
#
#         # ASM
#         finds = re.findall(
#             r"AUM</span>[^<]+<span[^>]+>([^<]+)</span>", html.text, re.DOTALL
#         )
#
#         # Check if the HTML contains only one occurrences.
#         if 0 == len(finds):
#             raise SymbolNotFound
#         if 1 < len(finds):
#             raise DataException(
#                 "More that one occurrences found in HTML - don't know what to do now"
#             )
#
#         data["Assets Under Management"] = finds[0]
#
#         # 52 week hi/low.
#         finds = re.findall(
#             r"52 Week Lo</span>[^<]+<span[^>]+>([^<]+)</span>", html.text, re.DOTALL
#         )
#
#         # Check if the HTML contains only one occurrences.
#         if 0 == len(finds):
#             raise SymbolNotFound
#         if 1 < len(finds):
#             raise DataException(
#                 "More that one occurrences found in HTML - don't know what to do now"
#             )
#
#         data["Year Low"] = finds[0]
#
#         finds = re.findall(
#             r"52 Week Hi</span>[^<]+<span[^>]+>([^<]+)</span>", html.text, re.DOTALL
#         )
#
#         # Check if the HTML contains only one occurrences.
#         if 0 == len(finds):
#             raise SymbolNotFound
#         if 1 < len(finds):
#             raise DataException(
#                 "More that one occurrences found in HTML - don't know what to do now"
#             )
#
#         data["Year High"] = finds[0]
#
#         return data
#
#     def get_holdings(self):
#         """
#         Returns ETF holdings (15 at max) list where each item is a list with items:
#
#         - symbol
#         - stock name
#         - percentage
#         """
#         try:
#             html = self._get(f"https://etfdb.com/etf/{self.symbol.upper()}/#holdings")
#         except Exception as e:
#             raise SymbolNotFound from e
#
#         finds = re.findall(
#             r"<table[^>]*etf-holdings[^>]*>(.*?)</table>", html.text, re.DOTALL
#         )
#
#         if not finds:
#             raise DataException("No holdings found.")
#
#         finds = re.findall(r"<tbody>(.*?)</tbody>", finds[0], re.DOTALL)
#         rows = []
#
#         if finds:
#             parser = HtmlTableParser(3)
#             parser.feed(finds[0])
#
#             rows = parser.get_data(sort_by=2)
#
#             return rows
#
#     def get_etfs_by_item(self):
#         """
#         Returns (top) 25 ETF's the item is included in
#         (sorted by it's weight).
#
#         Returned list contains of items where item is a list with items:
#
#         - symbol
#         - name
#         - category
#         - expense ratio
#         - weighting
#         """
#
#         try:
#             html = self._get(
#                 f"https://etfdb.com/stock/{self.symbol.upper()}/",
#                 follow_redirects=False,
#             )
#         except Exception as e:
#             raise SymbolNotFound from e
#
#         finds = re.findall(r"<table[^>]*>(.*?)</table>", html.text, re.DOTALL)
#
#         if not finds:
#             raise DataException("No ETFs found.")
#
#         finds = re.findall(r"<tbody>(.*?)</tbody>", finds[0], re.DOTALL)
#         rows = []
#
#         if finds:
#             parser = HtmlTableParser(5)
#             parser.feed(finds[0])
#
#             rows = parser.get_data()
#
#             return rows



import re

from .base import BaseAPI, HtmlTableParser
from .exceptions import DataException, SymbolNotFound


class StockAnalysis(BaseAPI):
    def get_etf_holdings(self):
        """
        Fetches ETF holdings table with following columns:

        - Symbol
        - Name
        - % Weight
        - Shares
        """

        try:
            html = self._get(
                f"https://stockanalysis.com/etf/{self.symbol.lower()}/holdings/"
            )
        except Exception as e:
            raise SymbolNotFound from e

        finds = re.findall(r"<table.*?>.*?</table>", html.text, re.DOTALL)

        # Check if the HTML contains only one table.
        if 0 == len(finds):
            raise SymbolNotFound
        if 1 < len(finds):
            raise DataException(
                "More that one table found in HTML - don't know what to do now"
            )

        parser = HtmlTableParser(columns=5)
        # monkey patch - fixing the case where table header is
        # improperly parsed  - "%" has it's own column and the
        # last "shares" column is then not parsed.
        parser.feed(finds[0])
        # parser.data[:6] = ("No.", "Symbol", "Name", "% Weight", "Shares")

        # Drop table header - first row.
        data = parser.get_data()[1:]
        # Drop first column - number of the row
        data = [i[1:] for i in data]

        return data

    # def get_current_price_change(self):
    #     """
    #     Fetches current market price inc. pre/post market
    #     prices/percent/value changes. Also returns current
    #     market state (pre-market, open, post-market, closed).
    #
    #     Fetched stucture has following fields:
    #
    #     - state (pre-market, open, post-market, closed)
    #     - pre_market
    #         - change
    #             - percents
    #             - value
    #         - value
    #     - current_market
    #         - change
    #             - percents
    #             - value
    #         - value
    #     - post_market
    #         - change
    #             - percents
    #             - value
    #         - value
    #
    #     Returned dict looks like:
    #
    #     .. code-block:: python
    #
    #         {
    #             "state": "open",
    #             "pre_market": {
    #                 "change": {
    #                     "percents": -1.32476,
    #                     "value": -1.42001
    #                 },
    #                 "value": 105.77
    #             },
    #             "current_market": {
    #                 "change": {
    #                     "percents": -1.6046284000000002,
    #                     "value": -1.7200012
    #                 },
    #                 "value": 105.47
    #             },
    #             "post_market": {
    #                 "change": {
    #                     "percents": 0.0,
    #                     "value": 0.0
    #                 },
    #                 "value": 0.0
    #             }
    #         }
    #
    #     :return: Current/Pre/Post market numbers (all are floats).
    #     :rtype: dict
    #     """
    #
    #     def get_state(data):
    #         """
    #         Returns one of following market states:
    #
    #         - open
    #         - closed
    #         - pre-market
    #         - post-market
    #         """
    #
    #         # Is state "extended" (meaning closed)?
    #         if not data["e"]:
    #             return data["ms"].lower()
    #         else:
    #             # Converts "After-hours" to "post-market".
    #             if "after-hours" == data["es"].lower():
    #                 return "post-market"
    #
    #             return data["es"].lower()
    #
    #     try:
    #         response = self._get(
    #             f"https://stockanalysis.com/api/quotes/s/{self.symbol.lower()}"
    #         )
    #     except Exception as e:
    #         raise SymbolNotFound from e
    #
    #     data = response.json()["data"]
    #     state = get_state(data)
    #     to_return = {
    #         "state": state,
    #     }
    #
    #     # Pre-market data.
    #     if "pre-market" == state:
    #         to_return["pre_market"] = {
    #             "change": {
    #                 "percents": data["ecp"],
    #                 "value": data["ec"],
    #             },
    #             "value": data["ep"],
    #         }
    #     else:
    #         to_return["pre_market"] = {
    #             "change": {
    #                 "percents": 0,
    #                 "value": 0,
    #             },
    #             "value": 0,
    #         }
    #
    #     # Market to_return.
    #     if "open" == data["ms"]:
    #         to_return["current_market"] = {
    #             "change": {
    #                 "percents": data["cp"],
    #                 "value": data["c"],
    #             },
    #             "value": data["p"],
    #         }
    #
    #     else:
    #         to_return["current_market"] = {
    #             "change": {
    #                 "percents": 0,
    #                 "value": 0,
    #             },
    #             "value": 0,
    #         }
    #
    #     # Post-market data.
    #     if "post-market" == state:
    #         to_return["post_market"] = {
    #             "change": {
    #                 "percents": data["ecp"],
    #                 "value": data["ec"],
    #             },
    #             "value": data["ep"],
    #         }
    #
    #     else:
    #         to_return["post_market"] = {
    #             "change": {
    #                 "percents": 0,
    #                 "value": 0,
    #             },
    #             "value": 0,
    #         }
    #
    #     return to_return



import re
from datetime import datetime

from .base import BaseAPI, HtmlTableParser
from .exceptions import SymbolNotFound


class FinViz(BaseAPI):
    """
    FinViz.com
    """

    def get_price_ratings(self):
        """
        Returns price ratings a.k.a price targets
        by analysts.

        Returned rows are:

        - date
        - status
        - analyst
        - rating
        - target price

        :return: Rows as a list of tuples where each tuple has 5 items.
        :rtype: list
        """

        try:
            html = self._get(
                f"https://finviz.com/quote.ashx?t={self.symbol.upper()}&ty=c&ta=1&p=d",
                headers={"User-Agent": self.user_agent},
            )
        except Exception as e:
            raise SymbolNotFound from e

        finds = re.findall(
            r"<table[^>]*js-table-ratings[^>]*>(.+?)</table>",
            html.text,
            re.DOTALL,
        )
        rows = []

        if finds:
            html = HtmlTableParser.fix_empty_cells(finds[0])
            parser = HtmlTableParser(columns=5)
            parser.feed(html)
            rows = parser.get_data()[1:]

        return rows

    def get_insider_trading(self):
        """
        Fetches insiders transactions (if available) as a
        list with following fields:

        - person
        - relationship
        - date
        - transaction
        - price
        - amount

        :return: Inriders transaction in reversed chronological order.
        :rtype: list
        """

        try:
            html = self._get(
                f"https://finviz.com/quote.ashx?t={self.symbol.upper()}&ty=c&ta=1&p=d",
                headers={"User-Agent": self.user_agent},
            )
        except Exception as e:
            raise SymbolNotFound from e

        rows = re.findall(r"<tr[^>]*insider-row*[^>]*>.+?<\/tr>", html.text, re.DOTALL)
        data = []

        if len(rows):
            parser = HtmlTableParser(columns=9)
            rows = "\n".join(rows)
            parser.feed(f"<table>{rows}</table>")

            for row in parser.get_data(sort_by=2):
                parsed_date = datetime.strptime(row[2], "%b %d '%y")
                data.append(
                    {
                        "person": row[0],
                        "relationship": row[1],
                        "date": parsed_date,
                        "transaction": row[3],
                        "price": float(row[4]),
                        "amount": int(row[5]),
                    }
                )

        return data



# import re
#
# import httpx
#
# from .base import BaseAPI
# from .exceptions import HttpException

# class Yahoo(BaseAPI):
#     """
#     Unofficial API wrapper class for TipRanks.com.
#     Unofficial means this class calls some hidden endpoints
#     and provides data that official API doesn't. Also doesn't
#     need an authorization.
#     """
#
#     def get_current_price_change(self):
#         """
#         Fetches current market price inc. pre/post market
#         prices/percent/value changes. Also returns current
#         market state (pre-market, open, post-market).
#
#         Fetched stucture has following fields:
#
#         - state (pre-market, open, post-market)
#         - pre_market
#             - change
#                 - percents
#                 - value
#             - value
#         - current_market
#             - change
#                 - percents
#                 - value
#             - value
#         - post_market
#             - change
#                 - percents
#                 - value
#             - value
#
#         Values are floats (if present) or None.
#         Returned dict looks like:
#
#         .. code-block:: python
#
#             {
#                 "state": "open",
#                 "pre_market": {
#                     "change": {
#                         "percents": -1.32476,
#                         "value": -1.42001
#                     },
#                     "value": 105.77
#                 },
#                 "current_market": {
#                     "change": {
#                         "percents": -1.6046284000000002,
#                         "value": -1.7200012
#                     },
#                     "value": 105.47
#                 },
#                 "post_market": {
#                     "change": {
#                         "percents": 0.0,
#                         "value": 0.0
#                     },
#                     "value": 0.0
#                 }
#             }
#
#         :return: Current/Pre/Post market numbers (all are floats or None).
#         :rtype: dict
#         """
#
#         def get_state(data):
#             if data["pre_market"]["value"]:
#                 return "pre-market"
#
#             if data["post_market"]["value"]:
#                 return "post-market"
#
#             return "open"
#
#         try:
#             response = self._get(
#                 f"https://finance.yahoo.com/quote/{self.symbol.upper()}/",
#                 headers={
#                     "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/117.0"
#                 },
#             )
#         except httpx.HTTPStatusError as e:
#             raise HttpException from e
#
#         finds = re.findall(r"<fin-streamer[^>].+?>", response.text, re.DOTALL)
#         need_map = {
#             "regularMarketPrice": "price",
#             "regularMarketChange": "change",
#             "regularMarketChangePercent": "change_percents",
#             "postMarketPrice": "post_market_price",
#             "postMarketChange": "post_market_change",
#             "postMarketChangePercent": "post_market_change_percents",
#             "preMarketPrice": "pre_market_price",
#             "preMarketChange": "pre_market_change",
#             "preMarketChangePercent": "pre_market_change_percents",
#         }
#         output = {}
#
#         if finds:
#             for f in finds:
#                 name = re.findall(r'data-field="([a-z]+)"', f, re.IGNORECASE)
#                 symbol = re.findall(r'data-symbol="([\^a-z]+)"', f, re.IGNORECASE)
#
#                 if (
#                     name[0] not in need_map.keys()
#                     or not symbol
#                     or self.symbol.upper() != symbol[0].upper()
#                 ):
#                     continue
#
#                 for key, val in need_map.items():
#                     if key == name[0]:
#                         value = re.findall(
#                             r'value="([-0-9\.a-z]+)"', f, re.DOTALL | re.IGNORECASE
#                         )
#
#                         output[val] = value[0]
#         output = {
#             "pre_market": {
#                 "change": {
#                     "percents": (
#                         float(val)
#                         if (val := output.get("pre_market_change_percents"))
#                         else None
#                     ),
#                     "value": (
#                         float(val) if (val := output.get("pre_market_change")) else None
#                     ),
#                 },
#                 "value": (
#                     float(val) if (val := output.get("pre_market_price")) else None
#                 ),
#             },
#             "current_market": {
#                 "change": {
#                     "percents": float(output["change_percents"]),
#                     "value": float(output["change"]),
#                 },
#                 "value": float(output["price"]),
#             },
#             "post_market": {
#                 "change": {
#                     "percents": (
#                         float(val)
#                         if (val := output.get("post_market_change_percents"))
#                         else None
#                     ),
#                     "value": (
#                         float(val)
#                         if (val := output.get("post_market_change"))
#                         else None
#                     ),
#                 },
#                 "value": (
#                     float(val) if (val := output.get("post_market_price")) else None
#                 ),
#             },
#         }
#         output["state"] = get_state(output)
#
#         return output



import logging
import re
from datetime import date, datetime
from html.parser import HTMLParser
from itertools import zip_longest

import httpx

from .exceptions import HttpException

logging.basicConfig(level=logging.DEBUG)


class BaseAPI:
    timeout = 10
    user_agent = (
        "Mozilla/5.0 (X11; Linux x86_64; rv:103.0) Gecko/20100101 Firefox/103.0"
    )

    def __init__(self, symbol=None):
        """
        Constructor.

        :param str symbol: Symbol of te item we wanna get info about.
        """

        if symbol:
            self.symbol = str(symbol)

    def _get(self, *args, **kwargs):
        """
        Wraps https.get() method and raises custom exception
        in case of httpx expcetion.
        Also rises an exception for any non 2xx or 3xx status.
        """

        kwargs.setdefault("follow_redirects", True)
        kwargs.setdefault("timeout", self.timeout)

        try:
            response = httpx.get(*args, **kwargs)
        except Exception as exc:
            raise HttpException(
                f"Couldn't perform GET request with args {args}"
            ) from exc

        response.raise_for_status()
        self.check_redirects(response)

        return response

    async def _aget(self, *args, **kwargs):
        kwargs.setdefault("follow_redirects", True)
        kwargs.setdefault("timeout", self.timeout)

        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(*args, **kwargs)
            except Exception as exc:
                raise HttpException(
                    f"Couldn't perform GET request with args {args}"
                ) from exc

            response.raise_for_status()
            self.check_redirects(response)

            return response

    def check_redirects(self, response):
        if response.url.path.startswith("/lookup"):
            raise HttpException(f"Couldn't find the symbol {self.symbol}")


class Data(dict):
    """
    Dict substitution which recursivelly handles
    non-existing keys.
    """

    def __getitem__(self, key):
        try:
            data = super().__getitem__(key)

            # If the data is dict we need to wrap it with
            # this class so it will carry this logic.
            if type(data) == dict:
                return self.__class__(data)

            # Data is not a dict so we return what we found.
            return data
        except:
            # In case of non existing key we return empty self
            # which makes sure another direct key demand will
            # copy this logic.
            return self.__class__()


class HtmlTableParser(HTMLParser):
    """
    Parses out all data from the given table and
    casts them into ``datetime.date`` or ``float``.

    Parsed data can be retrieved with ``get_data()`` method.
    """

    def __init__(self, columns, *args, **kwargs):
        """
        Constructor.

        :param int columns: Number of columns the given table has.
        """

        self.data = []
        self.in_cell = False
        self.cell_data = ""
        self.columns = columns
        super().__init__(*args, **kwargs)

    def handle_starttag(self, tag, attrs):
        if tag in ("td", "th"):
            self.in_cell = True
            self.cell_data = ""

    def handle_endtag(self, tag: str):
        if tag in ("td", "th"):
            self.in_cell = False
            self.data.append(self.cell_data)

    def handle_data(self, data):
        if self.in_cell:
            self.cell_data = self.parse_data(data)

    def parse_data(self, data):
        """
        Parses out all data from the given table and
        casts them into ``datetime.date`` or ``float``.
        """

        # logging.debug(f"data: {data}")

        # Date in YYYY-MM-DD format.
        if re.match(r"\d{4}-\d{2}-\d{2}", data):
            try:
                return date.fromisoformat(data)
            except:
                pass

        # Date in MM-DD-YY where MM is short string
        # representation - like "May" or "Apr"
        if re.match(r"^[a-zA-Z]{3}-\d{2}-\d{2}$", data):
            try:
                return datetime.strptime(data, "%b-%d-%y").date()
            except:
                pass

        # Date in MM/DD/YY format.
        if re.match(r"^\d{2}/\d{2}/\d{2}$", data):
            try:
                return datetime.strptime(data, "%m/%d/%y").date()
            except:
                pass

        if "today" == data.lower():
            return date.today()

        # Dollars (positive or negative floats).
        if re.match(r"^\$[+-]?([0-9]*[.])?[0-9]+$", data):
            try:
                return float(data[1:])
            except:
                pass

        # Int/float/percents (float or int with optional "%" sign as the last char).
        if re.match(r"^[0-9.,]+%?$", data):
            try:
                data = data.replace(",", "")
                return float(data[:-1] if "%" in data else data)
            except:
                pass

        if "--" == data:
            return 0.0

        return data

    def get_data(self, sort_by=0):
        """
        Splits data into ``self.columns`` list of lists
        and returns them.
        Rows are sorted chronologically.

        :param int sort_by: Column index to sort by. Use `None` to skip sorting.
        :return: Parsed, casted table data as rows.
        :rtype: list
        """

        data = list(zip_longest(*[iter(self.data)] * self.columns, fillvalue=""))

        if sort_by is not None:
            sorted_data = sorted(
                data[1:],
                key=lambda row: row[sort_by],
            )
            sorted_data.insert(0, data[0])

            return sorted_data
        return data

    @staticmethod
    def fix_empty_cells(html):
        """
        Fixes impty <td> cells and will substitude them with "--"
        which later gets recognized as 0.0 float.

        :param str html: HTML table to be fixed.
        :return: Fixed HTML.
        :rtype: str
        """
        return re.sub(r"<td([^>]*)><\/td>", "<td$1>--</td>", html, flags=re.DOTALL)


def strip_html_tags(text):
    """
    Removes HTML tags from the given string.

    :param str text: The string we want remove tags from.
    """

    return re.sub(r"<.*?>", "", text)


def calculate_perc_change(start, end):
    return (end - start) / start * 100



import asyncio
import time
from datetime import datetime

import httpx

from .base import BaseAPI, Data, calculate_perc_change
from .exceptions import HttpException, SymbolNotFound


class TipRanks(BaseAPI):
    """
    Unofficial API wrapper class for TipRanks.com.
    Unofficial means this class calls some hidden endpoints
    and provides data that official API doesn't. Also doesn't
    need an authorization.
    """

    def get_dividends(self):
        """
        Fetches symbol dividends with following fields:

        - yield
        - amount
        - ex_date
        - payment_date
        - record_date
        - growth_since

        :return: List of dividend objects.
        :rtype: list
        """

        try:
            response = self._get(
                f"https://tr-cdn.tipranks.com/bff/prod/stock/{self.symbol.lower()}/payload.json"
            )
        except httpx.HTTPStatusError as e:
            if 404 == e.response.status_code:
                raise SymbolNotFound
            raise HttpException from e

        data = response.json()

        dividends = []

        if data["dividend"]["history"]:
            for item in data["dividend"]["history"]:
                dividends.append(
                    {
                        "yield": float(item["yield"] or 0) * 100,
                        "amount": float(item["amount"]),
                        "ex_date": (
                            datetime.strptime(
                                item["executionDate"], "%Y-%m-%dT%H:%M:%S.000Z"
                            ).date()
                            if item["executionDate"]
                            else None
                        ),
                        "payment_date": (
                            datetime.strptime(
                                item["payDate"], "%Y-%m-%dT%H:%M:%S.000Z"
                            ).date()
                            if item["payDate"]
                            else None
                        ),
                        "record_date": (
                            datetime.strptime(
                                item["recordDate"], "%Y-%m-%dT%H:%M:%S.000Z"
                            ).date()
                            if item["recordDate"]
                            else None
                        ),
                        "growth_since": (
                            datetime.strptime(
                                item["growthSince"], "%Y-%m-%dT%H:%M:%S.000Z"
                            )
                            if item["growthSince"]
                            else None
                        ),
                    }
                )

        return dividends

    def get_basic_info(self):
        """
        Downloads basic info. Data are:

        - company_name
        - market
        - description
        - market_cap
        - has_dividends
        - yoy_change
        - year_low
        - year_high
        - pe_ratio
        - eps
        - similar_stocks
            - ticker
            - company_name

        :raises SymbolNotFound: In case data for the given symbol wasn't found.
        :return: Dict with data.
        :rtype: dict
        """

        async def download_basics(symbol):
            """
            Downloads basic info about symbol. Data are:

            - company_name
            - market
            - description
            - has_dividends
            - yoy_change
            - year_low
            - year_hign
            - pe_ratio
            - eps
            - market_cap
            - similar_stocks
                - ticker
                - company_name

            :param str symbol: Symbol the data will be downloaded for.
            :return: Dict with data.
            :rtype: dict
            """

            attempts = 5
            cool_off = 15
            i = 0
            json_data = None

            while i < attempts:
                try:
                    data = await self._aget(
                        f"https://www.tipranks.com/api/assets?tickers={symbol.upper()}",
                        headers={"User-Agent": self.user_agent, "Accept": "*/*"},
                    )
                except httpx.HTTPStatusError as e:
                    if 404 == e.response.status_code:
                        raise SymbolNotFound
                    raise HttpException from e

                try:
                    json_data = Data(data.json())
                    break
                except Exception:
                    print("Got nonsense response - trying again.")
                    i += 1
                    time.sleep(cool_off * i)

            if not json_data:
                return {}

            data = {
                "company_name": json_data["data"][0]["companyFullName"] or "",
                "market": json_data["extraData"][0]["research"]["stockMarketName"]
                or "",
                "description": json_data["extraData"][0]["research"]["description"]
                or "",
                "has_dividends": bool(json_data["data"][0]["dividendYield"] or None),
                "year_low": json_data["data"][0]["low52Weeks"] or 0.0,
                "year_high": json_data["data"][0]["high52Weeks"] or 0.0,
                "pe_ratio": json_data["data"][0]["peRatio"] or 0.0,
                "eps": json_data["data"][0]["lastReportedEps"]["reportedEPS"] or 0.0,
                "market_cap": json_data["data"][0]["marketCap"] or 0.0,
                "similar_stocks": [],
            }

            # YoY change.
            a_year_ago = json_data["data"][0]["landmarkPrices"]["yearAgo"]["p"]
            latest = json_data["extraData"][0]["prices"][-1]["p"]

            if a_year_ago and latest:
                data["yoy_change"] = float(calculate_perc_change(a_year_ago, latest))
            else:
                data["yoy_change"] = None

            return data

        async def download_similar(symbol):
            """
            Downloads similar stocks to the given symbol.

            :param str symbol: Symbol the data will be downloaded for.
            :return: Dict with data.
            :rtype: dict
            """

            try:
                data = await self._aget(
                    f"https://tr-cdn.tipranks.com/bff/prod/stock/{symbol.lower()}/payload.json",
                    headers={"User-Agent": self.user_agent},
                )
            except httpx.HTTPStatusError as e:
                # Checking if symbol was even found.
                if 404 == e.response.status_code:
                    raise SymbolNotFound
                raise HttpException from e

            json_data = Data(data.json())
            similar_stocks = []

            for item in json_data["similar"]["similar"][1:]:
                similar_stocks.append(
                    {"ticker": item["ticker"], "company_name": item["name"]}
                )

            return {"similar_stocks": similar_stocks}

        async def main():
            basic, additionals = await asyncio.gather(
                download_basics(self.symbol), download_similar(self.symbol)
            )
            basic.update(additionals)

            return basic

        return asyncio.run(main())

    def get_current_price_change(self):
        """
        Fetches current market price inc. pre/post market
        prices/percent/value changes. Also returns current
        market state (pre-market, open, post-market).

        Fetched stucture has following fields:

        - state (pre-market, open, post-market, closed)
        - pre_market
            - change
                - percents
                - value
            - value
        - current_market
            - change
                - percents
                - value
            - value
        - post_market
            - change
                - percents
                - value
            - value

        Values are floats (if present) or 0.0.
        Returned dict looks like:

        .. code-block:: python

            {
                "state": "open",
                "pre_market": {
                    "change": {
                        "percents": -1.32476,
                        "value": -1.42001
                    },
                    "value": 105.77
                },
                "current_market": {
                    "change": {
                        "percents": -1.6046284000000002,
                        "value": -1.7200012
                    },
                    "value": 105.47
                },
                "post_market": {
                    "change": {
                        "percents": 0.0,
                        "value": 0.0
                    },
                    "value": 0.0
                }
            }

        :return: Current/Pre/Post market numbers (all are floats).
        :rtype: dict
        """

        try:
            response = self._get(
                f"https://market.tipranks.com/api/quotes/GetQuotes?app_name=tr&tickers={self.symbol.upper()}"
            )
        except httpx.HTTPStatusError as e:
            raise HttpException from e

        data = response.json()["quotes"][0]
        output = {
            "state": "closed",
            "pre_market": {
                "change": {
                    "percents": 0.0,
                    "value": 0.0,
                },
                "value": 0.0,
            },
            "current_market": {
                "change": {
                    "percents": data["changePercent"],
                    "value": data["changeAmount"],
                },
                "value": data["price"],
            },
            "post_market": {"change": {"percents": 0.0, "value": 0.0}, "value": 0.0},
        }

        if data["isPremarket"]:
            output["pre_market"]["change"]["percents"] = data["prePostMarket"][
                "changePercent"
            ]
            output["pre_market"]["change"]["value"] = data["prePostMarket"][
                "changeAmount"
            ]
            output["pre_market"]["value"] = data["prePostMarket"]["price"]
            output["state"] = "pre-market"

        elif data["isAfterMarket"]:
            output["post_market"]["change"]["percents"] = data["prePostMarket"][
                "changePercent"
            ]
            output["post_market"]["change"]["value"] = data["prePostMarket"][
                "changeAmount"
            ]
            output["post_market"]["value"] = data["prePostMarket"]["price"]
            output["state"] = "post-market"

        elif data["isMarketOpen"]:
            output["state"] = "open"

        return output
